from msppy.msp import MSIP
from msppy.solver import SDDiP
from msppy.utils.examples import construct_nvic
from msppy.evaluation import Evaluation, EvaluationTrue
import numpy
import pandas as pd
import gurobipy
import multiprocessing




#data = pd.read_excel(r'/Users/jazzen/Documents/Pris.xlsx')
data = pd.read_excel(r'Pris.xlsx')
#t=19 equals 2040
#t=29 equals 2050

time = 30 #Equals 2051. We consider the problem until 2050
mu = 0.037  #Drift term in GBM
sigma = 0.29 #Diffusion term in GBM
#cap_fac = 0.58 #CAPACITY FACTOR FOR OW. Not used
tax = 0.78     #Petroleum tax on natural Gas
tax_1 = 0.22    # 22% corporate tax rate
alpha = 0.11 #Transmission loss in cables to shore


def sample_path_generator(random_state, size):    #Generates Markovian continous process
    output = numpy.zeros([size,time,7])
    output[:,0,1] = 417              #Initial Electricity price in 2021
    output[:,0,2] = 700              #Initial Carbon price in 2021
    output[:,0,3] = -output[:,0,1]*data['Adj-power'][0]*(1/(1-alpha)) + (1-tax)*data['Gas Price'][0]*data['GA-PFS'][0]\
                    + output[:,0,2]*data['CA-PFS'][0] + data['Others'][3]*data['NO-PFS'][0] - data['OM-PFS'][0]
    output[:,0,4] = (1-tax)*data['Gas Price'][0]*data['GA-OW'][0] + output[:,0,2]*data['CA-OW'][0] + data['Others'][3]*data['NO-OW'][0] - data['OM-OW'][0]
    output[:,0,5] = (1-tax_1)*(1-alpha)*data['OW-prod'][0]*output[:,0,1] - data['OM-OW2'][0]
    output[:,0,6] = data['Others'][4]*data['Adj-power'][0]*output[:,0,1] - output[:,0,4] - data['OM-OW'][0]
    for t in range(1,time):
        std_values = random_state.standard_normal(size=size)     #Generate random values
        output[:,t,0] = std_values
        output[:,t,1] = output[:,t-1,1]*numpy.exp((mu - 0.5*sigma**2) + sigma*output[:,t,0])      #Price according to GBM
        output[:,t,2] = data['Carbon1'][t]  #CO2 price
        output[:,t,3] = (-output[:,t,1]*data['Adj-power'][t]*(1/(1-alpha)) + (1-tax)*data['Gas Price'][t]*data['GA-PFS'][t]
                         + output[:,t,2]*data['CA-PFS'][t] + data['Others'][3]*data['NO-PFS'][t] - data['OM-PFS'][t])*data['Disc_fac'][t]  #PFS
        output[:,t,4] = ((1-tax)*data['Gas Price'][t]*data['GA-OW'][t] + output[:,t,2]*data['CA-OW'][t]
                         + data['Others'][3]*data['NO-OW'][t] - data['OM-OW'][t])*data['Disc_fac'][t]  #Offshore wind
        output[:,t,5] = ((1-tax_1)*(1-alpha)*data['OW-prod'][t]*output[:,t,1] - data['OM-OW2'][t])*data['Disc_fac'][t] #Exported power
        output[:,t,6] = (data['Others'][4]*data['Adj-power'][t]*output[:,t,1])*data['Disc_fac'][t]\
                        - output[:,t,4] - data['OM-OW'][t]*data['Disc_fac'][t]    #Cash flow terms for OW given PFS
    return output



nvm = MSIP(T = time, sense=-1, bound=100000000000000)       #Create the true problem
nvm.add_Markovian_uncertainty(sample_path_generator)

for t in range(time):
    model = nvm[t]
    #STATE VARIABLES and VARIABLES

    now_PFS, past_PFS = model.addStateVar(name='PFS', lb = 0, ub = 1, vtype='B', obj=1, uncertainty_dependent=3)
    now_OW, past_OW = model.addStateVar(name='OW', lb=0, ub=1, vtype='B', obj=1, uncertainty_dependent=4)
    now_EXP, past_EXP = model.addStateVar(name='EXP', lb=0, ub=1, vtype='B', obj=1, uncertainty_dependent=5)


    #NEW VARIABLES
    x_exp_ow = model.addVar(name='x_exp_ow', lb=0, ub=1, vtype='C')
    x_exp_pfs = model.addVar(name='x_exp_pfs', lb=0, ub=1, vtype='C')
    x_exp = model.addVar(name='x_exp', lb=0, ub=1, vtype='C')
    x_exp_pfs_ow = model.addVar(name='x_exp_pfs_ow', lb=0, ub=1, vtype='C')

    #Variables for OW given PFS
    x_ow_pfs_1 = model.addVar(name='x_ow_pfs_1', lb=0, ub=1, vtype='C')
    x_ow_pfs_2 = model.addVar(name='x_ow_pfs_2', lb=0, ub=1, vtype='C', obj=1, uncertainty_dependent=6)

    #Checks if no investment
    x_no_inv = model.addVar(name='no_inv', lb=0, ub=1, vtype='C', obj=0)

    #Variable for ssc in 2030
    x_ssc = model.addVar(name='ssc', lb=0, ub=1, vtype='C')


    #NEW CONSTRAINTS

    #Export if no prev decisions
    model.addConstr((1-now_EXP) + past_EXP + now_OW + now_PFS + x_exp >= 1)
    model.addConstr((1-x_exp) + now_EXP >= 1)
    model.addConstr(x_exp + past_EXP <= 1)
    model.addConstr(x_exp + now_OW <= 1)
    model.addConstr(x_exp + now_PFS <= 1)

    #Export if OW has been done previously
    model.addConstr((1-now_EXP) + past_EXP + (1-now_OW) + now_PFS + x_exp_ow >= 1)
    model.addConstr((1-x_exp_ow) + now_EXP >= 1)
    model.addConstr(x_exp_ow + past_EXP <= 1)
    model.addConstr((1-x_exp_ow) + now_OW >= 1)
    model.addConstr(x_exp_ow + now_PFS <= 1)

    #Export if PFS has been done previously
    model.addConstr((1-now_EXP) + past_EXP + now_OW + (1-now_PFS) + x_exp_pfs >= 1)
    model.addConstr((1-x_exp_pfs) + now_EXP >= 1)
    model.addConstr(x_exp_pfs + past_EXP <= 1)
    model.addConstr(x_exp_pfs + now_OW <= 1)
    model.addConstr((1-x_exp_pfs) + now_PFS >= 1)

    #Export if both OW + PFS has been done
    model.addConstr((1-now_EXP) + past_EXP + (1-now_OW) + (1-now_PFS) + x_exp_pfs_ow >= 1)
    model.addConstr((1-x_exp_pfs_ow) + now_EXP >= 1)
    model.addConstr(x_exp_pfs_ow + past_EXP <= 1)
    model.addConstr((1-x_exp_pfs_ow) + now_OW >= 1)
    model.addConstr((1-x_exp_pfs_ow) + now_PFS >= 1)

    #OW if PFS for investment cost
    model.addConstr((1-now_OW) + past_OW + (1-past_PFS) + now_EXP + x_ow_pfs_1 >= 1)
    model.addConstr((1-x_ow_pfs_1) + now_OW >= 1)
    #model.addConstr(x_ow_pfs_1 + past_OW <= 1)
    model.addConstr((1-x_ow_pfs_1) + past_PFS >= 1)
    model.addConstr((1-x_ow_pfs_1) + now_PFS >= 1)
    model.addConstr(x_ow_pfs_1 + now_EXP <= 1)



    #OW if PFS for cash flow
    model.addConstr((1-now_OW) + (1-past_PFS) + now_EXP + x_ow_pfs_2 >= 1)
    model.addConstr((1-x_ow_pfs_2) + now_OW >= 1)
    model.addConstr((1-x_ow_pfs_2) + past_PFS >= 1)
    model.addConstr(x_ow_pfs_2 + now_EXP <= 1)

    #Counts instances of no investment
    model.addConstr(now_OW + now_PFS + now_EXP + x_no_inv >= 1)
    model.addConstr(x_no_inv + now_OW <= 1)
    model.addConstr(x_no_inv + now_PFS <= 1)
    model.addConstr(x_no_inv + now_EXP <= 1)





    if t ==0:  #Ensures that past must be 0 in period 0. I.e. no investment is done previously
        model.addConstr(past_PFS==0)
        model.addConstr(past_OW==0)
        model.addConstr(past_EXP==0)


    #Ensures that once the decision is made, it remains done in all time periods

    model.addConstr(now_PFS >= past_PFS)
    model.addConstr(now_OW >= past_OW)
    model.addConstr(now_EXP >= past_EXP)


# INVESTMENT COST VARIABLES
    INV_PFS = model.addVar(name='Inv_PFS', lb=-100000000, ub=0, obj=1)
    model.addConstr(INV_PFS == (now_PFS-past_PFS)*(-1*data['Inv-PFS'][t])*data['Disc_fac'][t])

    INV_OW = model.addVar(name='Inv_OW', lb=-100000000, ub=0, obj=1)
    model.addConstr(INV_OW == (now_OW-past_OW)*-1*data['Inv-OW-SSC'][t]*data['Disc_fac'][t])


    Inv_exp = model.addVar(name='inv_exp', lb=-1000000000, ub=0, obj=1)
    model.addConstr(Inv_exp == x_exp*-data['Inv-EXP'][t]*data['Disc_fac'][t])


    Inv_exp_ow = model.addVar(name='inv_exp_ow', lb=-10000000000, ub=0, obj=1)
    model.addConstr(Inv_exp_ow == x_exp_ow*-data['Inv-EXP-OW'][t]*data['Disc_fac'][t])

    Inv_exp_pfs = model.addVar(name='inv_exp_pfs', lb=-1000000000, ub=0, obj=1)
    model.addConstr(Inv_exp_pfs == x_exp_pfs*-data['Inv-EXP-PFS'][t]*data['Disc_fac'][t])

    Inv_exp_pfs_ow = model.addVar(name='inv_exp_pfs_ow', lb=-1000000000, ub=0, obj=1)
    model.addConstr(Inv_exp_pfs_ow == x_exp_pfs_ow*-data['Inv-Both'][t]*data['Disc_fac'][t])

    Inv_ow_pfs = model.addVar(name='inv_ow_pfs', lb=-10000000000, ub=0, obj=1)
    model.addConstr(Inv_ow_pfs == x_ow_pfs_1*data['Others'][6]*data['Disc_fac'][t])

    price = model.addVar(name='El_Price', obj=0)
    model.addConstr(price == 0, uncertainty_dependent={'rhs': 1})
    
    #Constraints on time

    if t < 20:

        #Not allowing export before end of lifetime
        model.addConstr(now_EXP==0)
        model.addConstr(past_EXP==0)

    #Ensuring SSC cost in 2030
    if t == 9:
        SSC_inv = model.addVar(name='SSC', lb=-100000000, ub=0, obj=1)
        model.addConstr(SSC_inv == x_ssc*-data['Others'][2]*data['Disc_fac'][t])
        model.addConstr(now_OW + now_PFS + x_ssc >= 1)

#SOLVER
nvm.discretize(n_Markov_states=1, n_sample_paths=1000, method='SA', random_state=numpy.random.RandomState(6))
sddp = SDDiP(nvm)
sddp.solve(max_iterations=30, cuts=['SB'], n_simulations=1000, tol=0.01, freq_evaluations=3)
#Evaluates the problem
res_true = EvaluationTrue(nvm)
res_true.run(n_simulations=1000, percentile=90, query=['PFS', 'OW', 'EXP', 'x_exp_ow', 'x_exp_pfs', 'x_exp',
                                                       'x_exp_pfs_ow', 'x_ow_pfs_1', 'x_ow_pfs_2', 'El_Price', 'no_inv'])

print(res_true.CI)      #Confidence interval





#WRITING ANSWERS TO EXCEL
n_sim = 1000

d_PFS = numpy.zeros((n_sim, n_sim))
d_OW = numpy.zeros((n_sim, n_sim))
d_EXP = numpy.zeros((n_sim, n_sim))
d_exp_ow = numpy.zeros((n_sim, n_sim))
d_exp_pfs = numpy.zeros((n_sim, n_sim))
d_x_exp = numpy.zeros((n_sim, n_sim))
d_exp_pfs_ow = numpy.zeros((n_sim, n_sim))
d_ow_pfs = numpy.zeros((n_sim, n_sim))
d_elprice = numpy.zeros((n_sim, n_sim))
no_inv = numpy.zeros((n_sim, n_sim))
for t in range(30):
    d_PFS[t,:] = res_true.solution['PFS'][t]
    d_OW[t,:] = res_true.solution['OW'][t]
    d_EXP[t,:] = res_true.solution['EXP'][t]
    d_exp_ow[t,:] = res_true.solution['x_exp_ow'][t]
    d_exp_pfs[t,:] = res_true.solution['x_exp_pfs'][t]
    d_x_exp[t,:] = res_true.solution['x_exp'][t]
    d_exp_pfs_ow[t,:] = res_true.solution['x_exp_pfs_ow'][t]
    d_ow_pfs[t,:] = res_true.solution['x_ow_pfs_1'][t]
    d_elprice[t,:] = res_true.solution['El_Price'][t]
    no_inv[t,:] = res_true.solution['no_inv'][t]


df1 = pd.DataFrame(d_PFS)
df2 = pd.DataFrame(d_OW)
df3 = pd.DataFrame(d_EXP)
df4 = pd.DataFrame(d_exp_ow)
df5 = pd.DataFrame(d_exp_pfs)
df6 = pd.DataFrame(d_x_exp)
df7 = pd.DataFrame(d_exp_pfs_ow)
df8 = pd.DataFrame(d_ow_pfs)
df9 = pd.DataFrame(d_elprice)
df10 = pd.DataFrame(no_inv)

with pd.ExcelWriter('Test.xlsx') as writer:
    df1.to_excel(writer, sheet_name='PFS')
    df2.to_excel(writer, sheet_name='OW')
    df3.to_excel(writer, sheet_name='EXP')
    df4.to_excel(writer, sheet_name='exp_ow')
    df5.to_excel(writer, sheet_name='exp_pfs')
    df6.to_excel(writer, sheet_name='x_exp')
    df7.to_excel(writer, sheet_name='exp_pfs_ow')
    df8.to_excel(writer, sheet_name='ow_pfs')
    df9.to_excel(writer, sheet_name='elprice')
    df10.to_excel(writer, sheet_name='not_inv')

